import datetime

from .sws import Sws
""" This is the client.
    It needs a nice description.
"""
class SwsClient(Sws):

    def __init__(self, app_id, secret=None, user_id=0, timeout=3000, service_uri={}, auto_refresh=True, test_env=False):
        """
        Here we set up a mechanism for token refresh to be handled and triggered
        Create SWS object
        config : object
            Configuration options
        appId : str
            Application ID
        secret : str
            Application secret
        user_id : int
            End user ID
        timeout : float
            Request timeout
        service_uri : object
            Base URIs for SWS services
        service_uri.id : str
            Base URI for SWS ID Service
        service_uri.license : str
            Base URI for SWS License Service
        auto_refresh : boolean
            Determines if the client will attempt to refresh the access token if invalid or expired.
        test_env : boolean
            Determines if we want to run the unitest from travis.
        """
        super().__init__(app_id=app_id, secret=secret, user_id=user_id, timeout=timeout, service_uri=service_uri, invalid_access_token_handler=self._handle_invalid_access_token, test_env=test_env)
        self.auto_refresh = auto_refresh
        self.access_token_updated_callback = None

    def set_access_token_updated_callback(self, callback):
        """
        Sets a callback for when the access token has been updated when auto_refresh is true.
        :param function callback: Callback to execute when the access token has been updated.
        :return: Nothing.
        """
        self.access_token_updated_callback = callback

    def _handle_invalid_access_token(self, service, response):
        """
        Handles the case where a response came back from a service where the access token was not valid or was expired.

        If `auto_refresh` is true, will attempt to refresh the access token then attempt to make the last request again,
        otherwise returns the original response.

        :param Service service: The service which failed to fulfill a request due to an invalid access token.
        :param requests.Response response: The response given due to the invalid access token.
        :return: A HTTP response.
        :rtype: requests.Response
        """
        if self.auto_refresh:
            last_request = service.last_request
            response = self.identity().token_refresh(self.refresh_token)
            #   This token refresh request may have resulted in an error that was
            #   handled by a custom error handler.

            data = response.json()

            if 'tokens' in data:
                #   Update the access token property
                self.access_token = data['tokens']['access']['token']
                #   Call the callback
                if self.access_token_updated_callback:
                    self.access_token_updated_callback(
                      token=self.access_token,
                      expires=datetime.datetime.utcfromtimestamp(data['tokens']['access']['expires_at'])
                    )

                #   Set a new Authorization header for the request
                last_request.headers['Authorization'] = f'Bearer {self.access_token}'

                #   Re-execute the 'last request'
                return service.fetch_request(last_request)
        return response




import datetime
import hashlib
import random
import json

class FirewallHeader:
    def __init__(self):
        super().__init__()
    # Represents a strategy for generating a header that identifies Serato applications to the firewall. This header should
    # be non-trivial to guess by outsiders (unless they find this repository). If we wanted to make this header harder to
    # guess, we could introduce an environment variable that isn't present in the source code.

    # The values by which ordinal values of the ASCII characters in each 8-character chunk of the md5 hash will be
    # shifted

    SHIFTS = [-8, 8, -16, 16]


    # In another effort to make the header less guessable by people curious enough to make requests to our test servers
    # (but not curious enough to look at this open source code), add a prefix with characters drawn from a specific set
    # to the header

    PREFIX_CHARACTERS = 'serato'

    def getHeader(self):
        return {self.getHeaderKey(): self.getHeaderHash()}

    def getHeaderKey(self):
        return "x-serato-firewall"
        
    # Regular expression pattern that will match valid firewall header lines
    # Example match: "ras~/[Y*(0*Y9j:Ak8k9)T!Q )')@tFB@BDs"

    # HEADER_PATTERN = '/"[serato]{3}~[\x28-\x31\x59-\x5B\x5D\x5E\x79]{8}[\x38-\x41\x69-\x6E]{8}[\x20-\x21\x23-\x29\x51-\x56\x78]{8}[\x40-\x49\x71-\x76]{8}"/';


    # Returns a header value consisting of:
    # 1. A 3 character prefix drawn from characters in the PREFIX_CHARACTERS string with no repeats, and
    # 2. A hash, with every 8 ASCII character chunk shifted by the offsets defined in the SHIFTS array
    # - separated by a ~ character.
    # @return string Header value, for example "rta~Y[)(/*\,:ijkk>k:S!#R((U$tGvuIstE"

    def getHeaderHash(self):
        # @var string Date/time at which this header was created (used to create the hash)
        timeStamp = datetime.datetime.now().strftime("%d-%b-%Y (%H:%M:%S.%f)").encode('utf-8')
        # @var string Three letter prefix for the firewall header, from the set of PREFIX_CHARACTERS letters
        prefix = ''.join(random.choice(self.PREFIX_CHARACTERS) for i in range(3))

        g_hash = hashlib.md5(timeStamp).hexdigest()
        shiftedHash = ""
        for i in range(0,4):
            shiftedHash += self.shiftChunk(g_hash[i*8:i*8+8], self.SHIFTS[i])

        shiftedHash = self.replaceInvalidCharacters(shiftedHash)

        header = '\"' + prefix + '~' + shiftedHash + '\"'
        return header

    def shiftChunk(self, chunk, shift_num):
        shiftedChunk = ""
        for i in range(0,len(chunk)):
            shiftedChunk += self.shiftCharacter(chunk[i], shift_num)
        
        return shiftedChunk
        
    def shiftCharacter(self, character, shift_num):
        return chr(ord(character) + shift_num)

    # Characters to replace (and what to replace them with) in the header, since they'd be invalid in a quoted string
    # under RFC 2616 or RFC 7230

    def replaceInvalidCharacters(self, header):
        header = header.replace('\"','x')   
        header = header.replace('\\','y')
        return header


""" This class is a representation of the Serato Web Services.
    The purpose is to improve the integration of Python with our APIs.
    Encapsulates the functionality required to communicate with the endpoints we provide.
    The definitions of each service will be contained in separate classes
"""
from sws_py_sdk import identity, license, ecom

service_uri_default = {
    'id': 'id.serato.com',
    'license': 'license.serato.com',
    'ecom': 'ecom.serato.com',
    # profile: 'profile.serato.com'
}
class Sws(object):
    
    def __init__(self, app_id, secret=None, user_id=0, timeout=3000, service_uri={}, invalid_access_token_handler=None, test_env=False):
        """
        Create SWS object
        config : object
            Configuration options
        appId : str
            Application ID
        secret : str
            Application secret
        user_id : int
            End user ID
        timeout : float
            Request timeout
        service_uri : object
            Base URIs for SWS services
        service_uri.id : str
            Base URI for SWS ID Service
        service_uri.license : str
            Base URI for SWS License Service
        auto_refresh : boolean
            Determines if the client will attempt to refresh the access token if invalid or expired.
        """
        self.app_id = app_id
        self.secret = secret
        self.user_id = user_id
        self.timeout = timeout
        self.access_token = ''
        self.refresh_token = ''
        self.service_uris = {
            'id': service_uri['id'] if 'id' in service_uri.keys() else service_uri_default['id'],
            'license': service_uri['license'] if 'license' in service_uri.keys() else service_uri_default['license'],
            'ecom': service_uri['ecom'] if 'ecom' in service_uri.keys() else service_uri_default['ecom']
        }

        self.service = {
            'id': identity.Identity(sws=self),
            'license': license.License(sws=self),
            'ecom': ecom.Ecom(sws=self)
            # Define more clients here
        }
        self.invalid_access_token_handler = invalid_access_token_handler
        self.test_env = test_env

    def identity(self):
        """ Getter for the id service instance """
        return self.service['id']

    def license(self):
        """ Getter for the license service instance """
        return self.service['license']

    def ecom(self):
        """ Getter for the ecom service instance """
        return self.service['ecom']



""" This file exposes endpoints from the SWS Ecom Service
"""

from requests.auth import HTTPBasicAuth

from sws_py_sdk.service import Service

class Ecom(Service):

    def __init__(self, sws):
        super().__init__(sws)
        self.service_uri = sws.service_uris['ecom']

    def add_payment_method(self, nonce=None, device_data=None, billing_address_id=None):
        """ Add a payment method to the authenticated client user.
            nonce: str
                One-time-use reference to payment information provided by the user.
            device_data: str
                User device information.
            billing_address_id: str
                The two-letter value for an address associated with a specific customer ID.
        """
        return self.fetch(
            auth='bearer',
            endpoint='/api/v1/me/paymentmethods' if self.sws.user_id == 0 else '/api/v1/users/' +
                str(self.sws.user_id) + '/paymentmethods',
            body={"nonce": nonce, "device_data": device_data, "billing_address_id": billing_address_id },
            method='POST'
        )

    def add_plan_change_request(self, subscription_id, catalog_product_id):
        """ Create a new plan change that belongs to the authenticated client user.

            subscription_id: int
                Id of the subscription that is requesting change
            catalog_product_id: int
                The id of the product that the subscription wants to change to
        """
        endpoint = '/api/v1/me/' if self.sws.user_id == 0 else '/api/v1/users/' + str(self.sws.user_id)
        endpoint = endpoint + "subscriptions/" + subscription_id + "/planchanges"

        return self.fetch(
            auth='bearer',
            endpoint=endpoint,
            body={"catalog_product_id": catalog_product_id},
            method='POST'
        )

    def get_subscriptions(self, subscription_id=''):
        """ Get subscriptions owned by a user
        """
        endpoint = "/api/v1/me/subscriptions" if self.sws.user_id == 0 else "/api/v1/users/" + str(self.sws.user_id) + "/subscriptions"
        if subscription_id != '':
            endpoint += "/" + str(subscription_id)
        return self.fetch(
            auth="bearer",
            endpoint=endpoint,
            method="GET"
        )

    def get_invoice(self, order_id, invoice_id, accept='application/json'):
        """ Gets a specific invoice for an order

            order_id: int
                ID of the order
            invoice_id: int
                ID of the invoice
            accept: string
                Accept header, can be JSON, PDF or HTML
        """
        prefix = "/api/v1/me" if self.sws.user_id == 0 else "/api/v1/users/" + str(self.sws.user_id)
        endpoint = f"{prefix}/orders/{order_id}/invoices/{invoice_id}"
        return self.fetch(
            auth="bearer",
            endpoint=endpoint,
            method="GET",
            headers={'Accept': accept}
        )

    def get_orders(self, order_id=0):
        """ Get list of all orders created by the authenticated client user

            order_id: int
                ID of order
        """
        endpoint = "/api/v1/me" if self.sws.user_id == 0 else "/api/v1/users/" + str(self.sws.user_id)
        endpoint = endpoint + "/orders"
        if order_id != 0:
            endpoint = endpoint + "/" + str(order_id)
        return self.fetch(
            auth="bearer",
            endpoint=endpoint,
            method="GET"
        )

    def get_payment_methods(self):
        """ Gets users payment methods
        """
        endpoint = '/api/v1/me' if self.sws.user_id == 0 else '/api/v1/users/' + str(self.sws.user_id)
        endpoint += '/paymentmethods'
        
        return self.fetch(
            auth='bearer',
            endpoint=endpoint,
            method='GET'
        )
        

    def update_payment_methods(self, payment_token, nonce, device_data=None, billing_address_id=None):
        """ Update the payment method that belongs to the user

            payment_token: str
                Payment token string relating to user's payment method
            nonce: str
                One-time-use reference to payment information provided by the user.
            device_data: str
                User device information.
            billing_address_id: str
                The two-letter value for an address associated with a specific customer ID. 
        """
        endpoint = '/api/v1/me/paymentmethods/' if self.sws.user_id == 0 else '/api/v1/users/' + str(self.sws.user_id) + '/paymentmethods/'
        endpoint = endpoint + payment_token

        return self.fetch(
            auth='bearer',
            endpoint=endpoint,
            method='PUT',
            body={ "nonce": nonce, "device_data": device_data, "billing_address_id": billing_address_id }
        )
    def update_plan_change(self, subscription_id, plan_change_id):
        """ Update an existing plan change

            subscription_id: str
                Subscripton ID
            plan_change_id: str
                The ID for the change request
        """
        endpoint = '/api/v1/me' if self.sws.user_id == 0 else '/api/v1/users/' + str(self.sws.user_id)
        endpoint = endpoint + "/subscriptions/" + subscription_id + "/planchanges/" + plan_change_id
        return self.fetch(
            auth="bearer",
            endpoint=endpoint,
            method="PUT"
        )

    def update_subscription(self, subscription_id, number_of_billing_cycle=None, payment_method_token=None):
        """ Update a subscription owned by the user

            subscription_id: str
                Subscripton ID
            number_of_billing_cycle: int
                Number of billing cyles after which the subscription will expire.
            payment_method_token: str
                Token style identifier of the payment method
        """
        endpoint = '/api/v1/me' if self.sws.user_id == 0 else '/api/v1/users/' + str(self.sws.user_id)
        endpoint = endpoint + "/subscriptions/" + subscription_id
        return self.fetch(
            auth="bearer",
            endpoint=endpoint,
            body={"number_of_billing_cycle": number_of_billing_cycle, "payment_method_token": payment_method_token},
            method="PUT"
        )

    def delete_payment_method(self, payment_method_token):
        """ Deletes the user's payment method for the given payment method token

            payment_method_token: str
                Token style identifier of the payment method
        """
        endpoint = endpoint = '/api/v1/me' if self.sws.user_id == 0 else '/api/v1/users/' + str(self.sws.user_id)
        endpoint = endpoint + "/paymentmethods/" + payment_method_token
        return self.fetch(
            auth="bearer",
            endpoint=endpoint,
            method="DELETE"
        )

    def delete_subscription(self, subscription_id):
        """ Deletes the user's subscription for the given subscription id

            subscription_id: int
                Token style identifier of the payment method
        """
        endpoint = endpoint = '/api/v1/me' if self.sws.user_id == 0 else '/api/v1/users/' + str(self.sws.user_id)
        endpoint = endpoint + "/subscriptions/" + subscription_id
        return self.fetch(
            auth="bearer",
            endpoint=endpoint,
            method="DELETE"
        )
